from vidio.read import OpenCVReader
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np
import imageio
import tqdm
import h5py
import cv2
import os

na = np.newaxis

from .geometry import *
from .bundle_adjustment import embed_calib_objpoints


def pad_axis_limits(xmin, xmax, ymin, ymax, pad=0.1):
    """
    Pad x/y limits by a given fraction.

    Parameters
    ----------
    xmin, xmax, ymin, ymax : float
        Axis limits.

    pad : float, default=0.1
        Fraction of the axis range to pad.

    Returns
    -------
    xmin, xmax, ymin, ymax : float
        Padded axis limits.
    """
    x_range = xmax - xmin
    y_range = ymax - ymin
    xmin -= pad * x_range
    xmax += pad * x_range
    ymin -= pad * y_range
    ymax += pad * y_range
    return xmin, xmax, ymin, ymax


def set_axis_limits(ax, data, pctl=1, pad=0.1):
    """
    Set the axis limits of a matplotlib axis based on the data.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Axis to set limits for.

    data : array of shape (...,2)
        Data to use for setting the limits. NaNs are ignored.

    pctl : float, default=99
        Percentile of the data to use for setting the limits.

    pad : float, default=0.1
        Fraction of the axis range to pad.
    """
    xmin = np.nanpercentile(data[..., 0], pctl)
    xmax = np.nanpercentile(data[..., 0], 100 - pctl)
    ymin = np.nanpercentile(data[..., 1], pctl)
    ymax = np.nanpercentile(data[..., 1], 100 - pctl)
    xmin, xmax, ymin, ymax = pad_axis_limits(xmin, xmax, ymin, ymax, pad)
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)


def plot_residuals(
    all_calib_uvs,
    all_extrinsics,
    all_intrinsics,
    calib_objpoints,
    calib_poses,
    max_points=10000,
    marker_size=1,
    target_size=250,
    n_cols=3,
    inches_per_axis=5,
    hide_axes=True,
):
    """
    Visualize the reprojection error of the calibration.

    Estimated 3D coordinates of the calibration points (based on `calib_poses`)
    are projected into each camera's image plane using its extrinsic parameters.
    If calibration were perfect, the reprojections would be identical to the
    undistorted calibration detections (i.e. `cv2.undistortPoints` applied to
    `all_calib_uvs`). To visualize errors simultaneously across frames, a
    perspective transform is used to map the reprojection residuals into the
    canonical reference frame of the calibration object (i.e. `calib_objpoints`).

    A scatter plot is generated for each camera. The target (canonical)
    calibration points are shown as black "+" signs, and the reprojections
    are overlayed in red.

    Parameters
    ----------
    all_calib_uvs : array of shape (n_cameras, n_frames, N, 2)
        Calibration point detections for a number of frames for each camera.
        NaNs are used to indicate missing detections.

    all_extrinsics : array of shape (n_cameras, 6)
        Transforms from world coordinates to each camera's coordinates.

    all_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Camera intrinsics for each camera (see
        :py:func:`multicam_calibration.get_intrinsics`).

    calib_poses : array of shape (n_frames, 6)
        Calibration object poses in each frame.

    calib_objpoints : array of shape (N, 3)
        3D coordinates of the calibration points in a canonical reference frame.

    max_points : int, default=10000
        Maximum number of points to plot.

    marker_size : float, default=1
        Size of the markers used to plot the detections and reprojections.

    target_size : float, default=100
        Size of the markers used to plot the canonical calibration points.

    n_cols : int, default=3
        Number of columns in the figure.

    inches_per_axis : float, default=3
        Width of each axis in the figure (in inches).

    hide_axes : bool, default=True
        If True, the axes are hidden.

    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure containing the plot.

    median_error: array of shape (n_cameras,)
        Median distance between the reprojections and their target locations
        for each camera (in the same units as `calib_objpoints`).

    reprojections: array of shape (n_cameras, n_frames, N, 2)
        Reprojections of 3D calibration points into each camera's image plane
        in normalized (u,v) coordinates. These should ideally be identical to
        the detections (`all_calib_uvs`) after undisortion.

    transformed_reprojections: array of shape (n_cameras, n_frames, N, 2)
        Predicted coordinates of the calibration points transformed into the
        canonical reference frame of the calibration object. NaNs are used for
        camera/frame combinations where the calibration object was not visible.
    """

    n_cameras, n_frames, N, _ = all_calib_uvs.shape
    median_error = np.zeros(n_cameras)
    reprojections = np.zeros((n_cameras, n_frames, N, 2))
    transformed_reprojections = np.zeros((n_cameras, n_frames, N, 2)) * np.nan
    pts = embed_calib_objpoints(calib_objpoints, calib_poses)

    for cam in tqdm.trange(n_cameras):
        reprojections[cam] = project_points(
            pts, all_extrinsics[cam], all_intrinsics[cam][0]
        )
        uvs_undistorted = undistort_points(all_calib_uvs[cam], *all_intrinsics[cam])
        valid_ixs = np.nonzero(~np.isnan(uvs_undistorted).any((-1, -2)))[0]
        for t in valid_ixs:
            H = cv2.findHomography(uvs_undistorted[t], calib_objpoints[:, :2])
            transformed_reprojections[cam, t] = cv2.perspectiveTransform(
                reprojections[cam, t][na], H[0]
            )[0]

        errors = np.linalg.norm(
            transformed_reprojections[cam, valid_ixs] - calib_objpoints[:, :2],
            axis=-1,
        )
        median_error[cam] = np.median(errors)

    n_rows = int(np.ceil(n_cameras / n_cols))
    fig, axes = plt.subplots(n_rows, n_cols)
    for cam in range(n_cameras):
        pts = transformed_reprojections[cam].reshape(-1, 2)
        plot_ixs = np.nonzero(~np.isnan(pts).any(-1))[0]
        if len(plot_ixs) > max_points:
            plot_ixs = np.random.choice(plot_ixs, max_points, replace=False)

        axes.flat[cam].scatter(
            *calib_objpoints[:, :2].T,
            c="k",
            s=target_size,
            marker="+",
            linewidth=0.5,
        )
        axes.flat[cam].scatter(*pts[plot_ixs].T, c="r", s=marker_size, linewidth=0)
        axes.flat[cam].set_title(
            f"camera {cam} (median error={median_error[cam]:.2f})", fontsize=10
        )
        axes.flat[cam].set_aspect("equal")
        set_axis_limits(axes.flat[cam], pts[plot_ixs], pctl=1, pad=0.1)
        if hide_axes:
            axes.flat[cam].axis("off")

    aspect_ratio = calib_objpoints[:, 1].ptp() / calib_objpoints[:, 0].ptp()
    fig.set_size_inches(
        (n_cols * inches_per_axis, n_rows * inches_per_axis * aspect_ratio)
    )
    return fig, median_error, reprojections, transformed_reprojections


def overlay_detections(
    video_path,
    detections=None,
    output_path=None,
    overwrite=False,
    dotsize=3,
    draw_lines=True,
    frame_label_color=(0, 0, 255),
    frame_label_size=1,
    quality=6,
):
    """
    Overlay calibration object detections on a video.

    Parameters
    ----------
    video_path : str
        Path to a video of the calibration object.

    detections : dict, optional
        Dictionary mapping frame indexes to calibration object detections as
        arrays of shape (N,2). If not provided, the detections will be loaded
        from `{video_path}.detections.h5` (see
        :py:func:`multicam_calibration.detection.process_video` for the file
        format.

    output_path : str, optional
        Path to save the output video. If not provided, the video will be saved
        to `{video_path}.overlay.mp4`.

    overwrite : bool, default=False
        If True, overwrite the output video if it already exists.

    dotsize : int, default=3
        Size of the dots used to plot the calibration object points.

    draw_lines : bool, default=True
        If True, lines will be drawn between the calibration object points.

    frame_label_color : tuple (r,g,b), default=(0,0,255)
        Color of the frame labels.

    frame_label_size : int, default=1
        Font size of the frame labels.

    quality : int, default=6
        Quality of the output video (passed to `imageio.get_writer`)
    """
    if output_path is None:
        output_path = os.path.splitext(video_path)[0] + ".overlay.mp4"

    if os.path.exists(output_path) and not overwrite:
        print(f'{output_path} already exists. Set "overwrite=True" to overwrite.')
        return

    if detections is None:
        detections_path = os.path.splitext(video_path)[0] + ".detections.h5"

        assert os.path.exists(
            detections_path
        ), f"No detections found at {detections_path}."

        with h5py.File(detections_path, "r") as f:
            uvs = f["uvs"][()]
            frame_ixs = f["frame_ixs"][()]
            detections = {t: uvs[i] for i, t in enumerate(frame_ixs)}

    if output_path is None:
        output_path = os.path.splitext(video_path)[0] + ".overlay.mp4"

    reader = OpenCVReader(video_path)

    with imageio.get_writer(
        output_path, pixelformat="yuv420p", fps=reader.fps, quality=quality
    ) as writer:
        for frame_ix, frame in tqdm.tqdm(
            enumerate(reader), total=len(reader), ncols=72, unit="frame"
        ):
            if frame_ix in detections:
                uvs = detections[frame_ix]
                positions = [(int(u), int(v)) for u, v in uvs]
                colors = plt.cm.jet(np.linspace(0, 1, len(uvs)))[:, :3] * 255
                colors = [(int(r), int(g), int(b)) for r, g, b in colors]

                for pos, color in zip(positions, colors):
                    frame = cv2.circle(
                        frame, pos, dotsize, color, -1, lineType=cv2.LINE_AA
                    )

                if draw_lines:
                    for pos1, pos2, color in zip(positions[:-1], positions[1:], colors):
                        frame = cv2.line(
                            frame, pos1, pos2, color, 2, lineType=cv2.LINE_AA
                        )

            frame = cv2.putText(
                frame,
                str(frame_ix),
                (20, frame.shape[0] - 20),
                cv2.FONT_HERSHEY_SIMPLEX,
                frame_label_size,
                frame_label_color,
                2,  # thickness
                cv2.LINE_AA,
            )
            writer.append_data(frame)


def visualize_flatibration(
    transform, floor_points, keypoints=None, max_points_to_plot=5000, figsize=(12, 6)
):
    """
    Visualize the output of flatibration.

    Parameters
    ----------
    transform : array of shape (6,)
        Rigid transform output by :py:func:`multicam_calibration.flatibration.flatibrate`.

    floor_points : np.array of shape (n_points, 3) or a list of np.arrays
        3D keypoints that correspond to the floor of the recording arena.

    keypoints : np.array of shape (n_frames, n_keypoints, 3) or a list of np.arrays, optional
        3D keypoints of the animal(s) in the recording arena that were used for flatibration.

    max_points_to_plot : int, default=10000
        Maximum number of points to plot. If the number of points exceeds this value, a random
        subset of points will be plotted.

    figsize : tuple of int, default=(12, 12)
        Size of the figure.
    """
    fig, axs = plt.subplots(2, 2, figsize=figsize)

    if isinstance(floor_points, list):
        floor_points = np.concatenate(floor_points)

    if len(floor_points) > max_points_to_plot:
        ix = np.random.choice(len(floor_points), max_points_to_plot, replace=False)
        floor_points = floor_points[ix]

    # plot untransformed floor points
    axs[0, 0].scatter(*floor_points[:, [0, 2]].T, s=1, label="floor points", zorder=1)
    set_axis_limits(axs[0, 0], floor_points[:, [0, 2]], pctl=1, pad=0.1)
    axs[1, 0].scatter(*floor_points[:, [1, 2]].T, s=1, label="floor points", zorder=1)
    set_axis_limits(axs[1, 0], floor_points[:, [1, 2]], pctl=1, pad=0.1)

    # plot transformed floor points
    floor_points = apply_rigid_transform(transform, floor_points)
    axs[0, 1].scatter(*floor_points[:, [0, 2]].T, s=1, label="floor points", zorder=1)
    set_axis_limits(axs[0, 1], floor_points[:, [0, 2]], pctl=1, pad=0.1)
    axs[1, 1].scatter(*floor_points[:, [1, 2]].T, s=1, label="floor points", zorder=1)
    set_axis_limits(axs[1, 1], floor_points[:, [1, 2]], pctl=1, pad=0.1)

    if keypoints is not None:
        if isinstance(keypoints, list):
            keypoints = np.concatenate(keypoints)

        keypoints = keypoints.reshape(-1, 3)
        if len(keypoints) > max_points_to_plot:
            ix = np.random.choice(len(keypoints), max_points_to_plot, replace=False)
            keypoints = keypoints[ix]

        # plot untransformed keypoints
        axs[0, 0].scatter(*keypoints[:, [0, 2]].T, s=1, label="keypoints", zorder=0)
        set_axis_limits(axs[0, 0], keypoints[:, [0, 2]], pctl=1, pad=0.1)
        axs[1, 0].scatter(*keypoints[:, [1, 2]].T, s=1, label="keypoints", zorder=0)
        set_axis_limits(axs[1, 0], keypoints[:, [1, 2]], pctl=1, pad=0.1)

        # plot transformed keypoints
        keypoints = apply_rigid_transform(transform, keypoints)
        axs[0, 1].scatter(*keypoints[:, [0, 2]].T, s=1, label="keypoints", zorder=0)
        set_axis_limits(axs[0, 1], keypoints[:, [0, 2]], pctl=1, pad=0.1)
        axs[1, 1].scatter(*keypoints[:, [1, 2]].T, s=1, label="keypoints", zorder=0)
        set_axis_limits(axs[1, 1], keypoints[:, [1, 2]], pctl=1, pad=0.1)

    for ax in axs[:, 0]:
        ax.set_title("untransformed")

    for ax in axs[:, 1]:
        ax.set_title("transformed")

    for ax in axs[0]:
        ax.set_xlabel("x")

    for ax in axs[1]:
        ax.set_xlabel("y")

    for ax in axs.flat:
        ax.set_ylabel("z")
        ax.axhline(0, color="k", lw=1, zorder=2)

    plt.tight_layout()
    return fig



import numpy as np
from scipy.sparse import lil_matrix
from scipy.optimize import least_squares
import warnings
from .geometry import *

na = np.newaxis


def embed_calib_objpoints(calib_objpoints, calib_poses):
    """
    Embed calibration object points into the world coordinate system.

    Parameters
    ----------
    calib_objpoints : array of shape (N, 3)
        3D coordinates of the calibration points in a canonical reference frame.

    calib_poses : array of shape (n_frames, 6)
        Poses of the calibration object in world coordinates.

    Returns
    -------
    calib_worldpoints : array of shape (n_frames, N, 3)
        Calibration object points in world coordinates.
    """
    pose_transforms = get_transformation_matrix(calib_poses)[:, na]
    objpoints = euclidean_to_homogenous(calib_objpoints)[na, :, :, na]
    calib_worldpoints = (pose_transforms @ objpoints)[..., :3, 0]
    return calib_worldpoints


def predict_calib_uvs(
    all_extrinsics, all_intrinsics, calib_objpoints, calib_poses
):
    """
    Predict the (u,v) coordinates of calibration object points in each frame
    for each camera.

    Parameters
    ----------
    all_extrinsics : array of shape (n_cameras, 6)
        Transforms from world coordinates to each camera's coordinates.

    all_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Camera intrinsics for each camera (see
        :py:func:`multicam_calibration.get_intrinsics`).

    calib_objpoints : array of shape (N, 3)
        3D coordinates of the calibration points in a canonical reference frame.

    calib_poses : array of shape (n_frames, N, 3)
        Initial calibration object poses in each frame.

    Returns
    -------
    predicted_uvs : array of shape (n_cameras, n_frames, N, 2)
        Predicted (u,v) coordinates of calibration object points in each frame
        for each camera.
    """
    pts = embed_calib_objpoints(calib_objpoints, calib_poses)
    predicted_uvs = []
    for transform, intrinsics in zip(all_extrinsics, all_intrinsics):
        predicted_uvs.append(project_points(pts, transform, *intrinsics))
    return np.stack(predicted_uvs)


def residuals(params, all_calib_uvs, calib_objpoints):
    """
    Compute residuals between observed vs. expected (u,v) coordinates for a set
    of points based on their 3D positions and camera parameters.

    Parameters
    ----------
    params : array of shape (n_cameras * 12 + n_frames * 6)
        Camera parameters and calibration object poses in a single flat array.
        The camera parameters are `fx, fy, cx, cy, k1, k2, rx, ry, rz, tx, ty, tz`
        where `fx, fy` are focal lengths, `cx, cy` are principal points, `k1, k2`
        are radial distortion coefficients, `rx, ry, rz` define a 3D rotation in
        axis-angle form and `tx, ty, tz` define a translation. The calibration
        poses are rigid transforms (expressed in 6-dof vector form) from the
        calibration object's canonical coordinates to world coordinates.

    all_calib_uvs : array of shape (n_cameras, n_frames, N, 2)
        Calibration point detections for all the frames represented in `params`.
        NaNs are used to indicate missing detections.

    calib_objpoints : array of shape (N, 3)
        3D coordinates of the calibration object's points in its canonical
    """
    all_extrinsics, all_intrinsics, calib_poses = deserialize_params(
        params, all_calib_uvs.shape[0]
    )

    predicted_uvs = predict_calib_uvs(
        all_extrinsics, all_intrinsics, calib_objpoints, calib_poses
    )

    residuals = (all_calib_uvs - predicted_uvs)[~np.isnan(all_calib_uvs)]
    return residuals


def bundle_adjustment_sparsity(all_calib_uvs):
    """
    Create a sparse matrix representing the structure of the Jacobian matrix for
    bundle adjustment.

    Returns
    -------
    A : scipy.sparse.lil_matrix
        Sparse matrix representing the nonzero elements of the Jacobian matrix
        of the :py:func:`multicam_calibration.residuals` function.
    """
    n_cameras, n_frames, N, _ = all_calib_uvs.shape
    n_params = n_cameras * 12 + n_frames * 6

    mask = ~np.isnan(all_calib_uvs)
    camera_ixs = np.tile(
        np.arange(n_cameras)[:, na, na, na], (1, n_frames, N, 2)
    )[mask]
    frame_ixs = np.tile(
        np.arange(n_frames)[na, :, na, na], (n_cameras, 1, N, 2)
    )[mask]

    A = lil_matrix((mask.sum(), n_params), dtype=int)
    i = np.arange(mask.sum())
    for s in range(12):
        A[i, camera_ixs * 12 + s] = 1
    for s in range(6):
        A[i, n_cameras * 12 + frame_ixs * 6 + s] = 1
    return A


def serialize_params(all_extrinsics, all_intrinsics, calib_poses):
    """
    Serialize camera parameters and 3D calibration point positions into a flat
    array.

    Parameters
    ----------
    all_extrinsics : array of shape (n_camera, 6)
        Transforms from world coordinates to the coordinate system of each
        camera (see :py:func:`multicam_calibration.estimate_all_extrinsics`).

    all_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Camera intrinsics for each camera (see
        :py:func:`multicam_calibration.get_intrinsics`).

    calib_poses : array of shape (n_frames, 6)
        Rigid transforms (expressed in 6-dof vector form) from the calibration
        object's canonical coordinates to world coordinates.
    """
    x0 = []
    for transform, (camera_matrix, dist_coefs) in zip(
        all_extrinsics, all_intrinsics
    ):
        fx, fy, cx, cy = (
            camera_matrix[0, 0],
            camera_matrix[1, 1],
            camera_matrix[0, 2],
            camera_matrix[1, 2],
        )
        x0.append([fx, fy, cx, cy, *dist_coefs[:2], *transform])
    x0.append(calib_poses.flatten())
    return np.concatenate(x0)


def deserialize_params(x, n_cameras):
    """
    Deserialize camera parameters and calibration poses from a flat array
    (inverts :py:func:`multicam_calibration.serialize_params`).

    Parameters
    ----------
    x : array of shape (n_params,)
        Flat array of camera parameters and 3D calibration point positions.

    n_cameras : int
        Number of cameras.

    Returns
    -------
    all_extrinsics, all_intrinsics, calib_poses
        See :py:func:`multicam_calibration.serialize_params` for details.
    """
    all_extrinsics = []
    all_intrinsics = []
    for i in range(n_cameras):
        fx, fy, cx, cy = x[i * 12 : i * 12 + 4]
        camera_matrix = np.eye(3)
        camera_matrix[0, 0] = fx
        camera_matrix[1, 1] = fy
        camera_matrix[0, 2] = cx
        camera_matrix[1, 2] = cy
        dist_coefs = np.pad(x[i * 12 + 4 : i * 12 + 6], (0, 3))
        transform = x[i * 12 + 6 : i * 12 + 12]
        all_extrinsics.append(transform)
        all_intrinsics.append((camera_matrix, dist_coefs))
    calib_poses = x[n_cameras * 12 :].reshape((-1, 6))
    return np.array(all_extrinsics), all_intrinsics, calib_poses


def bundle_adjust(
    all_calib_uvs,
    all_extrinsics,
    all_intrinsics,
    calib_objpoints,
    calib_poses,
    n_frames=10000,
    outlier_pctl=95,
    **opt_kwargs
):
    """
    Bundle adjustment for camera parameters and 3D points.

    Camera parameters (focal length, principal point, distortion coefficients)
    and calibration object poses are all optimized simultaneously to minimize
    reprojection error, defined as the sum of squared residuals between observed
    and predicted (u,v) coordinates of calibration points. Frames with < 2
    cameras observing the calibration object are ignored.

    Parameters
    ----------
    all_calib_uvs : array of shape (n_cameras, n_frames, N, 2)
        Calibration point detections for a number of frames for each camera.
        NaNs are used to indicate missing detections.

    all_extrinsics : array of shape (n_cameras, 6)
        Transforms from world coordinates to each camera's coordinates.

    all_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Camera intrinsics for each camera (see
        :py:func:`multicam_calibration.get_intrinsics`).

    calib_objpoints : array of shape (N, 3)
        3D coordinates of the calibration points in a canonical reference frame.

    calib_poses : array of shape (n_frames, N, 3)
        Initial calibration object poses in each frame.

    n_frames : int, default=1000
        Number of randomly-sampled frames to use for bundle adjustment.
        If `None`, all frames are used.

    outlier_pctl : float, default=95
        Frames with reprojection error (max across cameras) greater than this
        percentile are excluded from bundle adjustment.

    opt_kwargs : dict
        Additional keyword arguments to pass to `scipy.optimize.least_squares`.
        The following arguments are set by default::

            verbose=2, x_scale='jac', ftol=1e-4, method='trf', loss='soft_l1'

    Returns
    -------
    adjusted_extrinsics : array of shape (n_cameras, 6)
        Adjusted camera extrinsics.

    adjusted_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Adjusted camera intrinsics.

    adjusted_calib_poses : array of shape (n_frames, 6)
        Adjusted calibration object poses.

    use_frames : array of shape (n_frames,)
        Indexes of the frames used in bundle adjustment.

    result : scipy.optimize.OptimizeResult
        Result object returned by `scipy.optimize.least_squares`.
    """
    n_cameras, _, N, _ = all_calib_uvs.shape
    use_frames = np.nonzero(
        (~np.isnan(all_calib_uvs).any((-1, -2))).sum(0) > 1
    )[0]

    # remove frames with high reprojection error
    predicted_uvs = predict_calib_uvs(
        all_extrinsics,
        all_intrinsics,
        calib_objpoints,
        calib_poses[use_frames],
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        sq_err = (all_calib_uvs[:, use_frames] - predicted_uvs) ** 2
        mean_sq_err = np.nanmax(np.nanmean(sq_err, axis=(-1, -2)), axis=0)

    outlier_threshold = np.nanpercentile(mean_sq_err, outlier_pctl)
    exclude = np.nan_to_num(mean_sq_err) > outlier_threshold
    use_frames = use_frames[~exclude]

    # sample frames
    if n_frames is None or n_frames > len(use_frames):
        n_frames = len(use_frames)
    else:
        use_frames = np.random.choice(use_frames, n_frames, replace=False)

    # setup optimization
    A = bundle_adjustment_sparsity(all_calib_uvs[:, use_frames])
    x0 = serialize_params(
        all_extrinsics, all_intrinsics, calib_poses[use_frames]
    )
    default_kwargs = dict(
        verbose=2, x_scale="jac", ftol=1e-4, method="trf", loss="soft_l1"
    )
    default_kwargs.update(opt_kwargs)

    # run optimization
    result = least_squares(
        residuals,
        x0,
        jac_sparsity=A,
        **default_kwargs,
        args=(all_calib_uvs[:, use_frames], calib_objpoints)
    )

    (
        adjusted_extrinsics,
        adjusted_intrinsics,
        adjusted_calib_poses,
    ) = deserialize_params(result.x, n_cameras)

    return (
        adjusted_extrinsics,
        adjusted_intrinsics,
        adjusted_calib_poses,
        use_frames,
        result,
    )



from vidio.read import OpenCVReader
from multiprocessing import Queue, Process
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tqdm
import h5py
import cv2
import os

from .geometry import euclidean_to_homogenous, homogeneous_to_euclidean

# -----------------------------------------------------------------------------#
#               General calibration object detection utilities                #
# -----------------------------------------------------------------------------#


def _worker(frame_queue, result_queue, detection_fun, detection_kwargs):
    detections = {}
    while True:
        item = frame_queue.get()
        if item == "STOP":
            break
        frame_ix, frame = item
        detection = detection_fun(frame, **detection_kwargs)
        if detection is not None:
            detections[frame_ix] = detection
    result_queue.put(detections)


def process_video(
    video_path, detection_fun, detection_options, use_frames=None, n_workers=1
):
    """
    Detect a calibration object in a video and save the detections.

    Detections are saved to an .h5 file with path `{video_path}.detections.h5`.
    The file contains the following datasets:

    - `uvs` : array of shape (n_frames, N, 2)
        Detected (u,v) coordinates of the calibration object points.

    - `qc_data` : array of shape (n_frames, M)
        Quality control information. The exact shape depends on the detection
        function used.

    - `frames_ixs` : array of shape (n_frames,)
        Index of the frame corresponding to each detection.

    - `img_size` : tuple (width, height)
        Image dimensions.

    Parameters
    ----------
    video_path : str
        Path to the video to process.

    detection_fun : callable
        Function to detect the calibration object in a single frame. Should
        take an image as input and return an array of (u,v) coordinates of the
        calibration object points.

    detection_options : dict
        Keyword arguments to pass to `detection_fun`.

    use_frames : array of shape (n_frames,), optional
        Indices of frames to use. If not provided, all frames will be used.

    n_workers : int, default=1
        Number of processes to use for multiprocessing.
    """
    reader = OpenCVReader(video_path)
    img_size = reader[0].shape[:2][::-1]
    total_frames = len(reader)

    if use_frames is None:
        use_frames = np.arange(total_frames)

    if n_workers > 1:
        frame_queue = Queue(maxsize=10)
        result_queue = Queue()

        processes = []
        for i in range(n_workers):
            p = Process(
                target=_worker,
                args=(
                    frame_queue,
                    result_queue,
                    detection_fun,
                    detection_options,
                ),
            )
            p.start()
            processes.append(p)

        try:
            for frame_ix in tqdm.tqdm(use_frames, ncols=72, unit="frame"):
                frame_queue.put((frame_ix, reader[frame_ix]))
        except KeyboardInterrupt:
            print("Early termination, saving results so far...")

        for _ in range(n_workers):
            frame_queue.put("STOP")

        detections = {}
        for _ in range(n_workers):
            detections.update(result_queue.get())

        for p in processes:
            p.join()

    else:
        detections = {}
        try:
            for frame_ix in tqdm.tqdm(use_frames, ncols=72, unit="frame"):
                frame = reader[frame_ix]
                detection = detection_fun(frame, **detection_options)
                if detection is not None:
                    detections[frame_ix] = detection
        except KeyboardInterrupt:
            print("Early termination, saving results so far...")

    if len(detections) == 0:
        raise ValueError("No detections found")
        
    
    frame_ixs = np.array(sorted(detections.keys()))
    if len(detections[frame_ixs[0]]) == 2:
        uvs = np.stack([detections[i][0] for i in frame_ixs])
        qc_data = np.stack([detections[i][1] for i in frame_ixs])
    else:
        uvs = np.stack([detections[i] for i in frame_ixs])
        qc_data = None

        
    save_path = os.path.splitext(video_path)[0] + ".detections.h5"
    with h5py.File(save_path, "w") as h5:
        h5.create_dataset("uvs", data=uvs)
        h5.create_dataset("frame_ixs", data=frame_ixs)
        h5.create_dataset("img_size", data=img_size)
        if qc_data is not None:
            h5.create_dataset("qc_data", data=qc_data)


def run_calibration_detection(
    video_paths,
    detection_fun,
    detection_options={},
    aligned_frame_ixs=None,
    overwrite=False,
    n_workers=1,
):
    """
    Detect a calibration object in a set of videos.

    Parameters
    ----------
    video_paths : list of str
        Paths to the videos to process.

    detection_fun : callable
        Function to detect the calibration object in a single frame. Should
        take an image as input and return an array of (u,v) coordinates of the
        calibration object points.

    detection_options : dict, optional
        Keyword arguments to pass to the detection function.

    aligned_frame_ixs : ndarray of shape (n_framesets, n_cameras), optional
        Indices of frames that are aligned across cameras. For example, if
        `aligned_frame_ixs[0,:2]==[0,1]` then frame 0 of the first camera is
        simultaneous with frame 1 of the second camera. NaNs indicate dropped
        frames. If `aligned_frame_ixs` is not provided, then the videos are
        assumed to be synchronized from the start and not lack dropped frames.

    overwrite : bool, default=False
        If True, ignore any saved detections and re-run the detection function
        for all frames.

    n_workers : int, default=1
        Number of processes to use for parallel processing.

    Returns
    -------
    all_calib_uvs : array of shape (n_cameras, n_frames, N, 2)
        Synchronized calibration detections each camera. NaNs are used when no
        detection was made or the frame was dropped.

    img_sizes : list of tuples (width, height)
        Image dimensions for each camera.
    """
    all_uvs = []
    all_img_sizes = []
    all_frame_ixs = []

    if aligned_frame_ixs is None:
        n_vids = len(video_paths)
        n_frames = min(len(OpenCVReader(vp)) for vp in video_paths)
        aligned_frame_ixs = np.stack([np.arange(n_frames)] * n_vids).T

    for i, video_path in enumerate(video_paths):
        save_path = os.path.splitext(video_path)[0] + ".detections.h5"
        if overwrite or not os.path.exists(save_path):
            print(f"Processing {video_path}")
            process_video(
                video_path,
                detection_fun,
                detection_options,
                aligned_frame_ixs[:, i],
                n_workers,
            )

        with h5py.File(save_path, "r") as h5:
            all_uvs.append(h5["uvs"][:])
            all_img_sizes.append(h5["img_size"][:])
            all_frame_ixs.append(h5["frame_ixs"][:])

    assert np.all(
        [len(uvs) > 0 for uvs in all_uvs]
    ), "At least one video has no detections"

    n_calib_points = all_uvs[0].shape[1]
    n_frames = aligned_frame_ixs.shape[0]
    n_cameras = len(video_paths)

    all_calib_uvs = np.zeros((n_cameras, n_frames, n_calib_points, 2)) * np.nan
    for i, (uvs, frame_ixs) in enumerate(zip(all_uvs, all_frame_ixs)):
        uvs = uvs[np.argsort(frame_ixs)]
        frame_ixs = np.sort(frame_ixs)
        all_calib_uvs[i, aligned_frame_ixs[:, i].searchsorted(frame_ixs)] = uvs

    return all_calib_uvs, all_img_sizes


def summarize_detections(all_calib_uvs):
    """
    Print the number of detections shared between each pair of cameras.

    Parameters
    ----------
    all_calib_uvs : array of shape (n_cameras, n_frames, N, 2)
        Synchronized calibration detections each camera. NaNs are used when no
        detection was made or the frame was dropped.

    Returns
    -------
    table : pandas.DataFrame
        Table of the number of shared detections between each pair of cameras.
    """
    has_detection = ~np.isnan(all_calib_uvs).any(axis=(2, 3))
    n_shared = (has_detection[:, None, :] & has_detection[None, :, :]).sum(2)
    names = [f"Camera {i}" for i in range(len(all_calib_uvs))]
    table = pd.DataFrame(n_shared, index=names, columns=names)
    return table


# -----------------------------------------------------------------------------#
#                            Chessboard detection                             #
# -----------------------------------------------------------------------------#


def extend_grid(uv_grid, extend_rows, extend_cols):
    """
    Given (u,v) coordinates from a chessboard detection in an image, simulate
    the coordinates of an enlarged chessboard using a homography transform.

    Parameters
    ----------
    uv_grid : array of shape (rows,cols,2)
        Grid of (u,v) coordinates (e.g. representing chessboard vertices).

    extend_rows : int
        Number of rows to extend by in each direction.

    extend_cols : int
        Number of columns to extend by in each direction.

    Returns
    -------
    extended_uv_grid : array of shape (rows+2*extend_rows,cols+2*extend_cols,2)
        Extended grid of (u,v) coordinates.
    """
    rows = uv_grid.shape[0] + 2 * extend_rows
    cols = uv_grid.shape[1] + 2 * extend_cols

    xy_grid_full = np.mgrid[0:cols, 0:rows].T
    xy_grid = xy_grid_full[extend_rows:-extend_rows, extend_cols:-extend_cols]

    H, _ = cv2.findHomography(xy_grid.reshape(-1, 2), uv_grid.reshape(-1, 2))

    extended_uv_grid = homogeneous_to_euclidean(
        euclidean_to_homogenous(xy_grid_full.reshape(-1, 2)) @ H.T
    ).reshape(xy_grid_full.shape)

    return extended_uv_grid


def detect_chessboard(
    image,
    *,
    board_shape=(7, 10),
    subpix_winSize=(5, 5),
    scale_factor=1,
    adaptive_threshold=True,
    normalize_image=True,
    reorder=True,
    match_score_min_diff=0.2,
):
    """
    Detect corners of a chessboard and order them using a anchor point.

    To use this function, the chessboard should have a dark circle in the
    top-left corner when viewed from the font. The dark circle serves as an
    anchor: Points are ordered such that the point nearest the dark circle is
    indexed first, followed by the others in row-major order.::

        ●
        ██░░██░░██
        ░░██░░██░░
        ██░░██░░██


    Parameters
    ----------
    image : array
        Image to detect the chessboard in.

    board_shape : tuple (rows,columns)
        Number of squares in each dimension minus one. For example the board
        shown above would have shape (2,4).

    subpix_winSize : tuple (width,height), default=(5,5)
        Size of the window to use for subpixel refinement.

    scale_factor : float, default=1
        How much to shrink the image before initial chessboard detection.
        In all cases, the full resolution image is subsequently used for
        subpixel refinement.

    adaptive_threshold : bool, default=True
        Whether to use the flat `CALIB_CB_ADAPTIVE_THRESH` when applying
        `findChessboardCorners` in OpenCV.

    normalize_image : bool, default=True
        Whether to use the flat `CALIB_CB_NORMALIZE_IMAGE` when applying
        `findChessboardCorners` in OpenCV.

    reorder : bool, default=True
        Whether to reorder the points using an anchor.

    match_score_min_diff : float, default=0.2
        Minimum difference between the best and second-best match score. If the
        difference is less than this, then the anchor point is ambiguous and
        the detection is rejected.


    Returns
    -------
    uvs: array of shape (N,2) or None
        Either an array of corner coordinates or None if no board was found.

    match_scores : array of shape (4,)
        Sorted template-matching correlations for the four possible anchor
        points locations. Only returned if `reorder=True`.
    """
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    flags = (
        cv2.CALIB_CB_FAST_CHECK
        + cv2.CALIB_CB_ADAPTIVE_THRESH * adaptive_threshold
        + cv2.CALIB_CB_NORMALIZE_IMAGE * normalize_image
    )
    if len(image.shape) == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    if scale_factor != 1:
        resized_image = cv2.resize(
            image, None, fx=scale_factor, fy=scale_factor
        )
    else:
        resized_image = image

    ret, corners_approx = cv2.findChessboardCorners(
        resized_image, board_shape, flags
    )

    if ret:
        corners_approx = corners_approx / scale_factor
        uvs = cv2.cornerSubPix(
            image, corners_approx, subpix_winSize, (-1, -1), criteria
        ).squeeze()

        if reorder:
            uvs, match_scores, _ = reorder_chessboard_corners(
                image, uvs, board_shape, 
            )
            if match_scores[0] - match_scores[1] < match_score_min_diff:
                return None
            else:
                return uvs, match_scores
        else:
            return uvs
    else:
        return None


def _generate_chessboard_anchor_template(size):
    """Generate a white square with a black circle in the top-left corner."""
    template = np.ones((size, size), dtype=np.uint8) * 255
    template = cv2.circle(template, (size // 4, size // 4), size // 4, 0, -1)
    return template


def _extract_region(image, source_pts, target_pts, template):
    """Extract a region of an image that will be matched to a template."""
    M = cv2.getPerspectiveTransform(source_pts, target_pts)
    warped_image = cv2.warpPerspective(image, M, template.shape[:2][::-1])
    return warped_image


def _match_to_template(image, all_source_pts, target_pts, template):
    """Compute correlation between a template and a region of an image."""
    match_scores, regions = [], []
    for source_pts in all_source_pts:
        region = _extract_region(image, source_pts, target_pts, template)
        if np.std(region) > 0:
            match_score = np.corrcoef(region.ravel(), template.ravel())[0, 1]
        else:
            match_score = 0
        match_scores.append(match_score)
        regions.append(region)
    return match_scores, regions


def reorder_chessboard_corners(image, uvs, board_shape, template_size=40):
    """
    Reorder chessboard points using an anchor point. See
    py:func:`multicam_calibration.detection.detect_chessboard` for details.

    Parameters
    ----------
    image : array
        Image that the chessboard was detected in.

    uvs : array of shape (N,2)
        Coordinates of the chessboard points.

    board_shape : tuple (rows,columns)
        Shape of the chessboard (number of squares in each dimension).

    template_size : int, default=40
        Size of template to generate.

    Returns
    -------
    reordered_uvs: array of shape (N,2)
        Reordered coordinates of the chessboard points.
    """
    if len(image.shape) == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

    uv_grid = uvs.reshape(board_shape[1], board_shape[0], 2)
    ext = extend_grid(uv_grid, 3, 1)

    all_source_pts = [
        np.float32([[ext[2, 0], ext[0, 0], ext[0, 2], ext[2, 2]]]),
        np.float32([[ext[0, -3], ext[0, -1], ext[2, -1], ext[2, -3]]]),
        np.float32([[ext[-3, -1], ext[-1, -1], ext[-1, -3], ext[-3, -3]]]),
        np.float32([[ext[-1, 2], ext[-1, 0], ext[-3, 0], ext[-3, 2]]]),
    ]

    s = 40  # size of anchor template
    template = _generate_chessboard_anchor_template(s)
    target_pts = np.float32([[0, s], [0, 0], [s, 0], [s, s]])
    match_scores, regions = _match_to_template(
        image, all_source_pts, target_pts, template
    )

    # Reorder points so best match is in top-left corner
    if np.argmax(match_scores) in [2, 3]:  # best match is in bottom row
        uv_grid = uv_grid[::-1, :]
    if np.argmax(match_scores) in [1, 2]:  # best match is in right column
        uv_grid = uv_grid[:, ::-1]

    uvs_reordered = uv_grid.reshape(-1, 2)
    sorted_match_scores = np.sort(match_scores)[::-1]
    vis_info = (all_source_pts, regions, template, match_scores)
    return uvs_reordered, sorted_match_scores, vis_info


def generate_chessboard_objpoints(chess_board_shape, chess_board_square_size):
    """
    Generate a list of 3D coordinates for the points on a chessboard
    in a standard orientation.

    The first point is assumed to be at (0,0,0) and the remaining points are
    spaced `square_size` apart in the x and y directions (with z=0).

    Parameters
    ----------
    chess_board_shape : tuple (rows,columns)
        Shape of the chessboard (number of squares in each dimension).

    square_size : float
        Size of a chessboard square (e.g. in mm).

    Returns
    -------
    objpoints : array of shape (N,3)
        Coordinates of the chessboard points.
    """
    rows, cols = chess_board_shape
    objpoints = np.zeros((rows * cols, 3), np.float32)
    objpoints[:, :2] = (
        np.mgrid[0:rows, 0:cols].T.reshape(-1, 2) * chess_board_square_size
    )
    return objpoints


def plot_chessboard_qc_data(video_paths, figsize=(12, 6)):
    """
    Plot the match scores that were used to reorder chessboard points in a set
    of videos.

    The top row contains heatmaps with sorted match scores from each frame. The
    bottom row contain scatter plots that compare the best and second-best
    match score on each frame (lines are offset from the diagonal at intervals
    of 0.1, and may be useful for selecting `match_score_min_diff` when running
    :py:func:`multicam_calibration.detect_chessboard`).

    Parameters
    ----------
    video_paths : list of str
        Paths to the videos to process. Each video should have a corresponding
        .detections.h5 file that contains a `qc_data` dataset.

    figsize : tuple (width,height), default=(12,6)
        Size of the figure.

    Returns
    -------
    fig : matplotlib Figure
        Figure containing the heatmaps.
    """
    qc_datas = []
    for video_path in video_paths:
        detections_path = os.path.splitext(video_path)[0] + ".detections.h5"
        assert os.path.exists(detections_path), (
            f"Could not find {detections_path}. "
            "Run `run_calibration_detection` first."
        )
        with h5py.File(detections_path, "r") as h5:
            qc_datas.append(h5["qc_data"][:])

    fig, axs = plt.subplots(
        3,
        len(qc_datas),
        figsize=figsize,
        gridspec_kw={"height_ratios": [1, 0.05, 0.4]},
    )

    for i, qc_data in enumerate(qc_datas):
        cax = axs[0, i].imshow(qc_data, aspect="auto", interpolation="nearest")

        axs[0, i].set_ylabel("Frame")
        axs[0, i].set_xlabel("match score")
        axs[0, i].set_title(os.path.basename(video_paths[i]), fontsize=10)
        plt.colorbar(cax, cax=axs[1, i], orientation="horizontal")

        axs[2, i].scatter(qc_data[:, 0], qc_data[:, 1], s=1, linewidth=0)
        axs[2, i].set_xlabel("best")
        axs[2, i].set_ylabel("second-best")
        axs[2, i].set_aspect("equal")
        axs[2, i].axline((0, 0), slope=1, c="k")
        axs[2, i].set_xlim([0, 1])
        axs[2, i].set_ylim([0, 1])

        for offset in np.arange(0.1, 1, 0.1):
            axs[2, i].axline((offset, 0), slope=1, c="k", ls="--", lw=0.5)

    plt.tight_layout()
    return fig



import numpy as np
import cv2
import tqdm

na = np.newaxis


def rodrigues(r):
    """
    Convert a rotation vector to a rotation matrix.

    Parameters
    ----------
    r : array of shape (...,3)
        Rotation vector in axis-angle form.

    Returns
    -------
    R : array of shape (...,3,3)
        Rotation matrix.
    """
    A = np.zeros((*r.shape[:-1], 3, 3))
    A[..., 0, 1] = -r[..., 2]
    A[..., 0, 2] = r[..., 1]
    A[..., 1, 0] = r[..., 2]
    A[..., 1, 2] = -r[..., 0]
    A[..., 2, 0] = -r[..., 1]
    A[..., 2, 1] = r[..., 0]
    theta = np.linalg.norm(r, axis=-1, keepdims=True).reshape(*r.shape[:-1], 1, 1)
    A = A / np.where(theta == 0, 1, theta)
    R = np.sin(theta) * A + (1 - np.cos(theta)) * np.matmul(A, A)
    R[..., 0, 0] += 1
    R[..., 1, 1] += 1
    R[..., 2, 2] += 1
    return R


def rodrigues_inv(R):
    """
    Convert a rotation matrix to a rotation vector.

    Parameters
    ----------
    R : array of shape (...,3,3)
        Rotation matrix.

    Returns
    -------
    r : array of shape (...,3)
        Rotation vector in axis-angle form.
    """
    r = np.stack(
        [
            R[..., 2, 1] - R[..., 1, 2],
            R[..., 0, 2] - R[..., 2, 0],
            R[..., 1, 0] - R[..., 0, 1],
        ],
        axis=-1,
    )
    theta = np.arccos((np.trace(R, axis1=-2, axis2=-1) - 1) / 2).reshape(
        *R.shape[:-2], 1
    )
    rnorm = np.linalg.norm(r, axis=-1, keepdims=True)
    rnorm += rnorm == 0
    return r * theta / rnorm


def rigid_transform_from_correspondences(source_points, target_points):
    """
    Compute the rigid transform that maps source points to target points.

    Parameters
    ----------
    source_points : array of shape (..., 3)
        Source points.

    target_points : array of shape (..., 3)
        Target points.

    Returns
    -------
    t : array of shape (6,)
        Rigid transforms in vector format. The first three elements specify a
        rotation in axis-angle form and the last three elements specify a
        translation.

    rmsd : float
        Root mean square deviation between the source and target points after
        applying the rigid transform.
    """
    # Reshape the points
    source_points = source_points.reshape(-1, 3)
    target_points = target_points.reshape(-1, 3)

    # Compute the centroids
    centroid_source = np.mean(source_points, axis=0)
    centroid_target = np.mean(target_points, axis=0)

    # Center the points around the centroids
    source_centered = source_points - centroid_source
    target_centered = target_points - centroid_target

    # Compute the covariance matrix H
    H = np.dot(source_centered.T, target_centered)

    # Perform Singular Value Decomposition
    U, S, Vt = np.linalg.svd(H)

    # Compute the rotation matrix
    R = np.dot(Vt.T, U.T)

    # Handle reflections
    if np.linalg.det(R) < 0:
        Vt[-1, :] *= -1
        R = np.dot(Vt.T, U.T)

    # Compute the translation vector and concatenate with the rotation vector
    translation = centroid_target - np.dot(R, centroid_source)
    t = np.concatenate((rodrigues_inv(R), translation))

    # Compute the root mean square deviation
    source_transformed = apply_rigid_transform(t, source_points)
    rmsd = np.sqrt(np.mean(np.sum((source_transformed - target_points) ** 2, axis=1)))

    return t, rmsd


def apply_rigid_transform(transform, points):
    """
    Apply a rigid transform to a set of points.

    Parameters
    ----------
    transform : array of shape (6,) or (4,4)
        Rigid transform in matrix or vector format. In vector format, the first three
        elements specify a rotation in axis-angle form and the last three elements
        specify a translation.

    points : array of shape (...,3)
        Points to transform.

    Returns
    -------
    points_transformed : array of shape (...,3)
        Transformed points.
    """
    if transform.shape == (6,):
        transform = get_transformation_matrix(transform)

    pts = euclidean_to_homogenous(points)
    points_transformed = np.matmul(transform, pts[..., na])[..., :3, 0]
    return points_transformed


def get_transformation_matrix(t):
    """
    Generate 4x4 transformation matrices from a 3D rotations and translations.

    Parameters
    ----------
    t : array of shape (...,6)
        Rigid transforms in vector format. The first three elements specify a
        rotation in axis-angle form and the last three elements specify a
        translation.

    Returns
    -------
    T : array of shape (...,4,4)
        Transformation matrices.
    """
    T = np.zeros((*t.shape[:-1], 4, 4))
    T[..., :3, :3] = rodrigues(t[..., :3])
    T[..., :3, 3] = t[..., 3:]
    T[..., 3, 3] = 1
    return T


def get_transformation_vector(T):
    """
    Convert 4x4 transformation matrices to vector format.

    This function inverts
    :py:func:`multicam_calibration.get_transformation_matrix`.

    Parameters
    ----------
    T : array of shape (...,4,4)
        Transformation matrices.

    Returns
    -------
    t : array of shape (...,6)
        Rigid transforms in vector format. The first three elements specify a
        rotation in axis-angle form and the last three elements specify a
        translation.
    """
    return np.concatenate([rodrigues_inv(T[..., :3, :3]), T[..., :3, 3]], axis=-1)


def get_projection_matrix(extrinsics, intrinsics):
    """
    Generate a projection matrix from camera intrinsics and extrinsics.

    The projection matrix transforms points from world coordinates to image
    coordinates. It is defined as follows, where K is the camera matrix, R is
    the rotation matrix, and t is the translation vector.::

        P = K [R | t]


    Parameters
    ----------
    extrinsics : array of shape (6,)
        The transformation from the world coordinate system to the camera's
        coordinate system. The first three elements are the rotation vector and
        the last three elements are the translation vector.

    intrinsics : tuple (camera_matrix, dist_coefs)
        Camera intrinsics (see :py:func:`multicam_calibration.get_intrinsics`).

    Returns
    -------
    P : array of shape (3, 4)
        The projection matrix.
    """
    camera_matrix, _ = intrinsics
    T = get_transformation_matrix(extrinsics)
    P = np.matmul(camera_matrix, T[:3])
    return P


def euclidean_to_homogenous(x_euclidean):
    """
    Convert Euclidean coordinates to homogenous coordinates

    ..math::

        (x_1,...,x_d) \mapsto (x_1,...,x_d, 1)

    Parameters
    ----------
    x_euclidean : array of shape (...,d)
        Euclidean coordinates.

    Returns
    -------
    x_homogenous : array of shape (...,d+1)
        Homogenous coordinates.
    """
    padding = np.ones((*x_euclidean.shape[:-1], 1))
    x_homogenous = np.concatenate((x_euclidean, padding), axis=-1)
    return x_homogenous


def homogeneous_to_euclidean(x_homogenous):
    """
    Convert homogenous coordinates to Euclidean coordinates

    ..math::

        (x_1,...,x_d, z) \mapsto (x_1/z,...,x_d/z)

    Parameters
    ----------
    x_homogenous : array of shape (...,d+1)
        Homogenous coordinates.

    Returns
    -------
    x_euclidean : array of shape (...,d)
        Euclidean coordinates.
    """
    x_euclidean = x_homogenous[..., :-1] / x_homogenous[..., -1:]
    return x_euclidean


def project_points(points, extrinsics, camera_matrix, dist_coefs=None):
    """
    Project 3D points onto the image plane.

    Parameters
    ----------
    points : array of shape (..., 3)
        3D points in the world coordinate system.

    extrinsics : array of shape (6,)
        The transformation from the world coordinate system to the camera's
        coordinate system. The first three elements are the rotation vector and
        the last three elements are the translation vector.

    camera_matrix : array of shape (3, 3)
        The camera matrix.

    dist_coefs : 1d array of length at least 2, optional
        The radial distortion coefficients k1, k2. If None, no distortion is
        applied.

    Returns
    -------
    uvs : array of shape (..., 2)
        The projected 2D points in the image plane.
    """
    # Rotate and translate points
    T = get_transformation_matrix(extrinsics)
    points_cam = apply_rigid_transform(T, points)

    # Apply radial distortion
    if dist_coefs is not None:
        k1, k2 = dist_coefs[:2]
        r2 = np.sum((points_cam[..., :2] / points_cam[..., 2:]) ** 2, axis=-1)
        radial_distortion = 1 + k1 * r2 + k2 * r2**2

        points_cam = points_cam * np.stack(
            (
                radial_distortion,
                radial_distortion,
                np.ones(points_cam.shape[:-1]),
            ),
            axis=-1,
        )

    # Project points to the image plane
    uvs = np.matmul(camera_matrix, points_cam[..., na]).squeeze(-1)
    uvs = uvs[..., :2] / uvs[..., 2:]
    return uvs


def undistort_points(uvs, camera_matrix, dist_coefs):
    """
    Wrapper for `cv2.undistortPoints` that handles NaNs and batch dimensions.

    Parameters
    ----------
    uvs : array of shape (..., 2)
        2D coordinates of points to undistort.

    camera_matrix : array of shape (3, 3)
        Camera matrix.

    dist_coefs : array of shape (5,)
        Distortion coefficients.

    Returns
    -------
    undistorted_uvs : array of shape (..., 2)
        Undistorted 2D coordinates.
    """
    uvs_shape = uvs.shape
    uvs = uvs.reshape(-1, 2)
    valid_ixs = ~np.isnan(uvs).any(-1)
    uvs_undistorted = np.zeros_like(uvs) * np.nan

    if valid_ixs.any():
        uvs_undistorted[valid_ixs] = cv2.undistortPoints(
            uvs[valid_ixs], camera_matrix, dist_coefs, None, camera_matrix
        ).squeeze(1)

    return uvs_undistorted.reshape(uvs_shape)


def triangulate(all_uvs, all_extrinsics, all_intrinsics):
    """
    Triangulate 3D points from 2D correspondences.

    This function performs robust triangulation by returning the median from
    all pairs of cameras that observe a given point.

    Parameters
    ----------
    all_uvs : list of arrays of shape (n_points, 2)
        2D coordinates of points in the image plane of each camera. Missing
        points should be represented by NaNs.

    all_extrinsics : all_extrinsics : list of arrays of shape (6,)
        Transforms from world to camera coordinates for each camera. The first
        three elements specify a rotation in axis-angle form and the last three
        elements specify a translation.

    all_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Camera intrinsics for each camera (see
        :py:func:`multicam_calibration.get_intrinsics`).

    Returns
    -------
    robust_points : array of shape (n_points, 3)
        Triangulated 3D points. NaNs indicate points that could not be
        triangulated (because they were observed by fewer than two cameras).
    """
    n_cameras = len(all_extrinsics)
    n_points = all_uvs[0].shape[0]

    # Undistort points
    all_undistorted_uvs = []
    for uvs, (camera_matrix, dist_coefs) in zip(all_uvs, all_intrinsics):
        all_undistorted_uvs.append(undistort_points(uvs, camera_matrix, dist_coefs))

    # Get projection matrices
    Ps = [
        get_projection_matrix(extrinsics, intrinsics)
        for extrinsics, intrinsics in zip(all_extrinsics, all_intrinsics)
    ]

    # Triangulate points from all pairs of cameras
    pairwise_points = []
    for i in range(n_cameras):
        for j in range(i + 1, n_cameras):
            points = np.zeros((n_points, 3)) * np.nan
            observed = ~np.any(
                [
                    np.isnan(all_undistorted_uvs[i]).any(-1),
                    np.isnan(all_undistorted_uvs[j]).any(-1),
                ],
                axis=0,
            )
            if np.any(observed):
                points_hom = cv2.triangulatePoints(
                    Ps[i],
                    Ps[j],
                    all_undistorted_uvs[i][observed].T,
                    all_undistorted_uvs[j][observed].T,
                ).T
                points[observed] = homogeneous_to_euclidean(points_hom)
            pairwise_points.append(points)
    pairwise_points = np.stack(pairwise_points, axis=0)

    # Take median of all triangulations
    robust_points = []
    for i in range(n_points):
        if np.isnan(pairwise_points[:, i]).all():
            robust_points.append(np.zeros(3) * np.nan)
        else:
            robust_points.append(np.nanmedian(pairwise_points[:, i], axis=0))
    return np.stack(robust_points, axis=0)



"""
The goal of flatibration is to compute a 3D rigid transform to a coordinate system where
the XY plane corresponds to the floor of the recording arena. This code implements the
following pipeline:

Input:
  - option 1: 3D keypoints from animals in the recording arena
  - option 2: 3D keypoints corresponding to the floor of the recording arena

Output:
    3D rigid transform that maps from the coordinate system of the input keypoints to 
    a coordinate system where the XY plane is the floor of the recording arena

Algorithm:
    Suppose z=ax+by+t is the equation of the plane identified by RANSAC. This means
    that vectors (1,0,a) and (0,1,b) are parallel to the floor plane, (a, b, -1) is 
    orthogonal to the floor plane. The output transformation is uniquely determined by:
      - map the centroid of the floor points to the origin
      - map (1,0,a) to the X-axis
      - map (0,1,b) to the Y-axis
      - map (-a, -b, 1) to the Z-axis
"""

import numpy as np
from sklearn.linear_model import RANSACRegressor

from .geometry import (
    rigid_transform_from_correspondences,
    get_transformation_matrix,
    get_transformation_vector,
)


def get_floor_points(keypoints, z_points_down=False):
    """
    Get the 3D keypoints that correspond to the floor of the recording arena.

    Parameters
    ----------
    keypoints : np.array of shape (n_frames, n_keypoints, 3) or a list of np.arrays
        3D keypoints of the animal(s) in the recording arena. The z-axis should be
        roughly vertical.

    z_points_down : bool, default=False
        True if the z-axis points down, False if the z-axis points up.
    """
    if isinstance(keypoints, list):
        keypoints = np.concatenate(keypoints)

    if z_points_down:
        ix = np.argmax(keypoints[:, :, 2], axis=1)
    else:
        ix = np.argmin(keypoints[:, :, 2], axis=1)
    return keypoints[np.arange(keypoints.shape[0]), ix]


def flatibrate(floor_points, residual_threshold=10):
    """
    Compute a 3D rigid transform that maps the floor points to the XY plane and moves
    the centroid of the floor points to the origin.

    Parameters
    ----------
    floor_points : np.array of shape (n_points, 3) or a list of np.arrays
        3D keypoints that correspond to the floor of the recording arena.

    residual_threshold : float, default=10
        Maximum residual for a data point to be considered an inlier (passed to RANSAC)

    Returns
    -------
    transform : np.array of shape (6,)
        Rigid transform in vector format. The first three elements specify a rotation in
        axis-angle form and the last three elements specify a translation.
    """
    if isinstance(floor_points, list):
        floor_points = np.concatenate(floor_points)

    ransac = RANSACRegressor(residual_threshold=residual_threshold).fit(
        floor_points[:, :2], floor_points[:, 2]
    )
    a, b = ransac.estimator_.coef_

    source_x_axis = np.array([1, 0, a])
    source_y_axis = np.array([0, 1, b])
    source_z_axis = np.array([-a, -b, 1])
    source_origin = np.median(floor_points, axis=0)

    source_pts = np.array(
        [
            source_origin,
            source_origin + source_x_axis,
            source_origin + source_y_axis,
            source_origin + source_z_axis,
        ]
    )

    target_pts = np.array(
        [
            [0, 0, 0],
            [np.linalg.norm(source_x_axis), 0, 0],
            [0, np.linalg.norm(source_y_axis), 0],
            [0, 0, np.linalg.norm(source_z_axis)],
        ]
    )

    transform, _ = rigid_transform_from_correspondences(source_pts, target_pts)
    return transform


def flip_z_axis(transform):
    """
    Compose the input rigid transform with a 180 degree rotation around the X-axis.

    Parameters
    ----------
    transform : np.array of shape (6,)
        Rigid transform in vector format. The first three elements specify a rotation in
        axis-angle form and the last three elements specify a translation.

    Returns
    -------
    transform_flipped : np.array of shape (6,)
        Rigid transform in vector format. The first three elements specify a rotation in
        axis-angle form and the last three elements specify a translation.
    """
    T = get_transformation_matrix(transform)
    T_flipped = np.diag([1, -1, -1, 1]) @ T
    transform_flipped = get_transformation_vector(T_flipped)
    return transform_flipped



from .bundle_adjustment import *
from .flatibration import *
from .calibration import *
from .detection import *
from .geometry import *
from .viz import *
from .io import *

from . import _version

__version__ = _version.get_versions()["version"]



import cv2, os, json
import numpy as np
import h5py

from .geometry import get_transformation_matrix, rodrigues_inv


def save_calibration(
    all_extrinsics, all_intrinsics, camera_names, save_path, save_format="json"
):
    """
    Save calibration results.

    Several formats are supported. In every case, the extrinsics are saved
    as 3x3 rotation matrices and 3x1 translation vectors that map from world
    coordinates to each camera's coordinates. The intrinsics are saved as
    3x3 camera matrices and distortion coefficients `k1, k2, p1, p2, k3`.
    The available formats include:

    - `json`: Single JSON file mapping camera names to calibration parameters.

    - `jarvis`: Directory with one file per camera in OpenCV-specific YAML format.
      The rotation and camera matrices are transposed relative to the JSON format.

    - `gimbal`: Single HDF5 file with a `camera_parameters` group containing
        `camera_names`, `dist_coefs`, `intrinsic`, `rotation`, and `translation`.

    Parameters
    ----------
    all_extrinsics : array of shape (n_cameras, 6)
        Transformations from the world coordinate system to each camera's
        coordinate system. The first three elements are the rotation vector
        and the last three elements are the translation vector.

    all_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Intrinsic parameters for each camera (see
        :py:func:`multicam_calibration.get_intrinsics`).

    camera_names : list of str
        Names of the cameras. Must be the same length as `all_extrinsics`

    save_path : str
        Path to save calibration. Interpreted as a file if `format=="json"`
        and as a directory if `format=="jarvis"`.

    save_format : str, default='json'
        Format to save calibration. See above for options.
    """
    assert (
        len(all_extrinsics) == len(all_intrinsics) == len(camera_names)
    ), "Number of camera names must match number of extrinsics and intrinsics"

    transforms = get_transformation_matrix(np.array(all_extrinsics))

    if save_format == "json":
        calibration_data = {}
        for i, camera_name in enumerate(camera_names):
            calibration_data[camera_name] = {
                "R": transforms[i, :3, :3].tolist(),
                "T": transforms[i, :3, 3:].tolist(),
                "camera_matrix": all_intrinsics[i][0].tolist(),
                "distortion_coefs": all_intrinsics[i][1].tolist(),
            }
        if not save_path.endswith(".json"):
            save_path += ".json"
        with open(save_path, "w") as f:
            json.dump(calibration_data, f, indent=4)

    elif save_format == "jarvis":
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        for i, camera_name in enumerate(camera_names):
            filename = os.path.join(save_path, f"{camera_name}.yaml")
            fs = cv2.FileStorage(filename, cv2.FILE_STORAGE_WRITE)
            fs.write("intrinsicMatrix", all_intrinsics[i][0].T)
            fs.write(
                "distortionCoefficients", all_intrinsics[i][1].reshape(1, -1)
            )
            fs.write("R", transforms[i, :3, :3].T)
            fs.write("T", transforms[i, :3, 3:])
            fs.release()

    elif save_format == "gimbal":
        camera_matrix = np.stack(
            [intrinsics[0] for intrinsics in all_intrinsics]
        )
        dist_coefs = np.stack([intrinsics[1] for intrinsics in all_intrinsics])
        if not save_path.endswith(".h5"):
            save_path += ".h5"
        with h5py.File(save_path, "w") as h5:
            grp = h5.create_group("camera_parameters")
            grp.create_dataset("dist_coefs", data=dist_coefs)
            grp.create_dataset("intrinsic", data=camera_matrix)
            grp.create_dataset("rotation", data=transforms[:, :3, :3])
            grp.create_dataset("translation", data=transforms[:, :3, 3])
            grp.create_dataset("camera_names", data=camera_names)

    else:
        raise ValueError(f"Unknown format {save_format}")


def load_calibration(load_path, load_format="json", camera_names=None):
    """
    Load calibration results.

    A range of formats are supported. In every case, the extrinsics are loaded
    as 6-vectors where the first three elements are a rotation vector and the
    last three elements are a translation vector. The intrinsics are loaded as
    3x3 camera matrices and 5x1 distortion coefficients. The available
    formats include

    - `json`: A single JSON file mapping camera names to calibration parameters.

    - `jarvis`: Directory with one file per camera in OpenCV-specific YAML format.
        The rotation and camera matrices are transposed relative to the JSON format.

    - `gimbal`: Single HDF5 file with a `camera_parameters` group containing
        `camera_names`, `dist_coefs`, `intrinsic`, `rotation`, and `translation`.

    Parameters
    ----------
    load_path : str
        Path to calibration. Interpreted as a file if `format=="json"`
        and as a directory if `format=="jarvis"`.

    camera_names : list of str
        Desired camera order. Can all be used to only load a subset of
        cameras. If `camera_names==None` the order will be alphabetical.

    load_format : str, default='json'
        Format to load calibration. See above for options.

    Returns
    -------
    all_extrinsics : list of arrays of shape (6,)
        Extrinsics for each camera (see above for format).

    all_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Intrinsic parameters for each camera (see above for format).

    camera_names : list of str
        Names of the cameras, in an order that matches `all_extrinsics` and
        `all_intrinsics`.
    """
    if load_format == "json":
        with open(load_path, "r") as f:
            calibration_data = json.load(f)
        if camera_names is None:
            camera_names = sorted(calibration_data.keys())
        else:
            assert set(camera_names) == set(
                calibration_data.keys()
            ), "Camera names must match keys in calibration file"

        all_extrinsics, all_intrinsics = [], []
        for camera_name in camera_names:
            all_extrinsics.append(
                np.concatenate(
                    [
                        rodrigues_inv(
                            np.array(calibration_data[camera_name]["rotation"])
                        ),
                        np.array(
                            calibration_data[camera_name]["translation"]
                        ).squeeze(),
                    ]
                )
            )
            all_intrinsics.append(
                (
                    np.array(calibration_data[camera_name]["camera_matrix"]),
                    np.array(
                        calibration_data[camera_name]["distortion_coefs"]
                    ),
                )
            )
        return all_extrinsics, all_intrinsics, camera_names

    elif load_format == "jarvis":
        yaml_files = [
            f
            for f in sorted(os.listdir(load_path))
            if os.path.splitext(f)[1] in [".yaml", ".YAML"]
        ]
        names_to_files = {os.path.splitext(f)[0]: f for f in yaml_files}
        if camera_names is None:
            camera_names = sorted(names_to_files.keys())
        else:
            assert set(camera_names) <= set(
                names_to_files.keys()
            ), "Camera names must be a subset of yaml files in calibration directory"

        all_extrinsics, all_intrinsics = [], []
        for camera_name in camera_names:
            filepath = os.path.join(load_path, names_to_files[camera_name])
            fs = cv2.FileStorage(filepath, cv2.FILE_STORAGE_READ)
            all_extrinsics.append(
                np.concatenate(
                    [
                        rodrigues_inv(fs.getNode("R").mat().T),
                        fs.getNode("T").mat().squeeze(),
                    ]
                )
            )
            all_intrinsics.append(
                (
                    fs.getNode("intrinsicMatrix").mat().T,
                    fs.getNode("distortionCoefficients").mat().squeeze(),
                )
            )
        return all_extrinsics, all_intrinsics, camera_names

    elif load_format == "gimbal":
        if not load_path.endswith(".h5"):
            load_path += ".h5"
        with h5py.File(load_path, "r") as h5:
            grp = h5["camera_parameters"]

            h5_names = [
                name.decode("utf-8")
                for name in grp["camera_names"][()].tolist()
            ]

            all_intrinsics = list(
                zip(grp["intrinsic"][()], grp["dist_coefs"][()])
            )

            all_extrinsics = np.concatenate(
                [rodrigues_inv(grp["rotation"][()]), grp["translation"][()]],
                axis=1,
            )

            if camera_names is None:
                camera_names = h5_names
            else:
                assert set(camera_names) <= set(
                    h5_names
                ), "Camera names must be a subset of names in calibration file"
                ix = np.array([h5_names.index(name) for name in camera_names])
                all_extrinsics = all_extrinsics[ix]
                all_intrinsics = [all_intrinsics[i] for i in ix]
            return list(all_extrinsics), all_intrinsics, camera_names

    else:
        raise ValueError(f"Unknown format {load_format}")



import setuptools

with open("README.md", 'r') as f:
    long_description = f.read()

setuptools.setup(
    name='multicam-calibration',
    version='0.0.0',
    author='Caleb Weinreb',
    author_email='calebsw@gmail.com',
    include_package_data=True,
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent'
    ],
    python_requires='>=3.7',
    install_requires=[
        'scipy',
        'vidio',
        'networkx',
        'tqdm',
        'matplotlib',
    ], 
    url='https://github.com/calebweinreb/multicam-calibration',
)



import numpy as np
import networkx as nx
import cv2
import tqdm
import warnings
from .geometry import *

na = np.newaxis


def get_intrinsics(
    calib_uvs,
    calib_objpoints,
    image_size,
    n_samples=100,
    fix_k3=True,
    zero_tangent_dist=True,
):
    """
    Get camera intrinsics from the (u,v) coordinates of calibration points
    detections in a set of images.

    Parameters
    ----------
    calib_uvs : array of shape (n_frames, N, 2)
        Calibration point detections for a number of frames.

    calib_objpoints : array of shape (N, 3)
        3D coordinates of the calibration points in a canonical reference frame.

    image_size : tuple (width, height)
        Size of the images in pixels.

    n_samples : int, default=100
        Number of samples to use for calibration.

    fix_k3 : bool, default=True
        If True, the third radial distortion coefficient is fixed to zero.

    zero_tangent_dist : bool, default=True
        If True, the tangential distortion coefficients are fixed to zero.

    Returns
    -------
    camera_matrix : array of shape (3,3)
        Camera matrix, which has the form::

            | fx  0   cx |
            | 0   fy  cy |
            | 0   0   1  |

    dist_coefs : array of shape (5,)
        Distortion coefficients (k1,  k2,  p1,  p2,  k3).
    """
    calib_uvs = calib_uvs[~np.isnan(calib_uvs).any((1, 2))]
    n_samples = min(n_samples, len(calib_uvs))

    imgpoints = [
        calib_uvs[i]
        for i in np.random.choice(len(calib_uvs), n_samples, replace=False)
    ]
    imgpoints = np.array(imgpoints).astype(np.float32)

    calib_objpoints = np.repeat(calib_objpoints[na], n_samples, axis=0).astype(
        np.float32
    )
    flags = (
        cv2.CALIB_FIX_K3 * fix_k3
        + cv2.CALIB_ZERO_TANGENT_DIST * zero_tangent_dist
    )

    camera_matrix, dist_coefs = cv2.calibrateCamera(
        calib_objpoints, imgpoints, image_size, None, None, flags=flags
    )[1:3]
    return camera_matrix, dist_coefs.squeeze()


def estimate_pose(calib_uvs, calib_objpoints, camera_matrix, dist_coeffs):
    """
    Estimate the pose of a calibration object in a series of frames.

    The pose in each frame is the rigid transform from the calibration
    object's canonical coordinate frame to the camera's coordinate frame.

    Parameters
    ----------
    calib_uvs : array of shape (n_frames, N, 2)
        Calibration point detections for a number of frames.

    calib_objpoints : array of shape (N, 3)
        3D coordinates of the calibration points in a canonical reference frame.

    camera_matrix : array of shape (3, 3)
        Camera matrix (use :py:func:`multicam_calibration.get_intrinsics` to
        calculate).

    dist_coeffs : array of shape (5,)
        Distortion coefficients (use
        :py:func:`multicam_calibration.get_intrinsics` to calculate).

    Returns
    -------
    poses : array of shape (n_frames, 6)
        Pose of the calibration object in each frame, where the first three
        columns are the rotation in axis-angle form and the last three columns
        are the translation. NaNs in a row indecate that the pose could not be
        estimated or that the calibration object was not detected in that frame.
    """
    poses = np.zeros((len(calib_uvs), 6)) * np.nan
    for ix, imgpoints in enumerate(calib_uvs):
        if not np.isnan(imgpoints).any():
            retval, rvec, tvec = cv2.solvePnP(
                calib_objpoints, imgpoints, camera_matrix, dist_coeffs
            )
            if retval:
                poses[ix] = np.concatenate([rvec.squeeze(), tvec.squeeze()])
    return poses


def estimate_pairwise_camera_transform(camera1_poses, camera2_poses):
    """
    Estimate the transformation between two cameras.

    Parameters
    ----------
    camera1_poses : array of shape (n_frames, 6)
        Calibration object poses of the first camera (see
        :py:func:`multicam_calibration.estimate_pose`).

    camera2_poses : array of shape (n_frames, 6)
        Calibration object poses of the second camera (see
        :py:func:`multicam_calibration.estimate_pose`).

    Returns
    -------
    transform : arrays of shape (6,)
        Transformation from the first camera's coordinate system to the second
        camera's coordinate system in vector format, where the first three
        elements specify a rotation in axis-angle form and the last three
        elements specify a translation.
    """
    common_frames = ~np.isnan([camera1_poses, camera2_poses]).any((0, 2))
    T1 = get_transformation_matrix(camera1_poses[common_frames])
    T2 = get_transformation_matrix(camera2_poses[common_frames])
    T_rel = np.matmul(T2, np.linalg.inv(T1))
    transforms = get_transformation_vector(T_rel)
    return np.median(transforms, axis=0)


def get_camera_spanning_tree(all_calib_poses, root=0):
    """
    Estimate the maximal spanning tree over camera pairs.

    This function outputs the maximal spanning tree in a graph where each node
    represents a camera and each edge represents the number of frames with 
    calibration object detections for both cameras.

    Parameters
    ----------
    all_calib_poses : array-like with shape (n_cameras, n_frames, 6)
        Calibration object poses across frames for each camera. Frames with no
        calibration object detections should be set to NaN (see 
        :py:func:`multicam_calibration.estimate_pose`).

    root : int, default=0
        Index of the root camera (used for ordering edges in the spanning tree).

    Returns
    -------
    spanning_tree : list of tuples
        Each tuple contains the indexes of two cameras that are connected by an
        edge in the spanning tree. The edges are topoligically sorted by 
        distance from the root, with the first node in each tuple being nearer
        to the root than the second node. For example, the tree::

            0
            |
            1
           / \
          2   3

        with root "0" would represented as::

            [(0,1), (1,2), (1,3)]
    """
    edges = []
    n_cameras = len(all_calib_poses)
    calib_detected = ~np.isnan(all_calib_poses).any(2)
    for i in range(n_cameras):
        for j in range(i + 1, n_cameras):
            n_common_frames = (
                calib_detected[i, :] & calib_detected[j, :]
            ).sum()
            edges.append((i, j, n_common_frames))

    G = nx.Graph()
    G.add_nodes_from(range(n_cameras))
    G.add_weighted_edges_from(edges)
    spanning_tree = nx.maximum_spanning_tree(G)
    root_dist = nx.shortest_path_length(spanning_tree, source=root)
    edges = [
        tuple(sorted(e, key=lambda n: root_dist[n]))
        for e in spanning_tree.edges
    ]
    edges = sorted(edges, key=lambda e: root_dist[e[0]])
    return edges


def estimate_all_extrinsics(all_calib_poses, root=0):
    """
    Estimate transformations from the root camera to all cameras.

    Parameters
    ----------
    all_calib_poses : array-like with shape (n_cameras, n_frames, 6)
        Calibration object poses across frames for each camera. Frames with no
        calibration object detections should be set to NaN (see
        :py:func:`multicam_calibration.estimate_pose`).

    root : int, default=0
        Index of the root camera.

    Returns
    -------
    all_extrinsics : array of shape (n_cameras, 6)
        Transforms from the root camera's coordinate system to the coordinate
        system of each camera (the root camera's transform is the identity).
        The first three elements specify a rotation in axis-angle form and the
        last three elements specify a translation.
    """
    all_extrinsics = [None] * len(all_calib_poses)
    all_extrinsics[root] = np.eye(4)
    spanning_tree = get_camera_spanning_tree(all_calib_poses, root=root)

    for c1, c2 in spanning_tree:
        transform = estimate_pairwise_camera_transform(
            all_calib_poses[c1], all_calib_poses[c2]
        )
        all_extrinsics[c2] = (
            get_transformation_matrix(transform) @ all_extrinsics[c1]
        )
    all_extrinsics = np.array(
        list(map(get_transformation_vector, all_extrinsics))
    )
    return all_extrinsics


def consensus_calib_poses(all_calib_poses, all_extrinsics):
    """
    Estimate the consensus pose of the calibration object in each frame.

    Poses estimated from each camera separately are mapped transformed into
    common (world) coordinates and then the median across cameras is used as
    the final estimate.

    Parameters
    ----------
    all_calib_poses : array-like with shape (n_cameras, n_frames, 6)
        Calibration object poses across frames for each camera. Frames with no
        calibration object detections should be set to NaN (see
        :py:func:`multicam_calibration.estimate_pose`).

    all_extrinsics : array of shape (n_camera, 6)
        Transforms from world coordinates to the coordinate system of each
        camera (see :py:func:`multicam_calibration.estimate_all_extrinsics`).

    Returns
    -------
    calib_poses : array of shape (n_frames, 6)
        Consensus calibration object pose in each frame. Frames with no
        calibration detections in any camera are set to NaN.
    """
    all_calib_poses_world_coords = np.zeros_like(all_calib_poses) * np.nan
    for i, (poses, transform) in enumerate(
        zip(all_calib_poses, all_extrinsics)
    ):
        calib_detected = ~np.isnan(poses).any(axis=-1)
        T_board2camera = get_transformation_matrix(poses[calib_detected])
        T_world2camera = get_transformation_matrix(transform)
        T_board2world = np.matmul(
            np.linalg.inv(T_world2camera), T_board2camera
        )
        all_calib_poses_world_coords[
            i, calib_detected
        ] = get_transformation_vector(T_board2world)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        calib_poses = np.nanmedian(all_calib_poses_world_coords, axis=0)
    return calib_poses


def calibrate(
    all_calib_uvs,
    img_sizes,
    calib_objpoints,
    root=0,
    verbose=True,
    n_samples_for_intrinsics=100,
):
    """
    Estimate 3D positions of calibration points across frames and
    the intrinsic and extrinsic parameters of each camera.

    Parameters
    ----------
    all_calib_uvs : array of shape (n_cameras, n_frames, N, 2)
        Calibration point detections for a number of frames for each camera.
        NaNs are used to indicate missing detections.

    img_sizes : list of tuples (width, height)
        Image sizes for each camera.

    calib_objpoints : array of shape (N, 3)
        3D coordinates of the calibration points in a canonical reference frame.

    root : int, default=0
        Index of the camera to use as the root of the coordinate system.

    verbose : bool, default=True
        If True, print progress updates.

    n_samples_for_intrinsics : int, default=100
        Number of frames to use for estimating camera intrinsics (pass as
        `n_samples` to :py:func:`multicam_calibration.get_intrinsics`).

    Returns
    -------
    all_extrinsics : array of shape (n_cameras, 6)
        Transforms from the root camera's coordinate system to the coordinate
        system of each camera (the root camera's transform is the identity).
        The first three elements specify a rotation in axis-angle form and the
        last three elements specify a translation.

    all_intrinsics : list of tuples (camera_matrix, dist_coefs)
        Camera intrinsics for each camera (see
        :py:func:`multicam_calibration.get_intrinsics`).

    consensus_calib_poses : array of shape (n_frames, 6)
        Calibration object poses in world coordinates, defined as the rigid
        transformation from the canonical reference frame of the calibration
        object to world coordinates. Frames with no calibration object
        detections in any camera are set to NaN.
    """
    n_cameras = len(all_calib_uvs)

    all_intrinsics = []
    if verbose:
        print("Estimating camera intrinsics")
    for uvs, img_size in tqdm.tqdm(
        zip(all_calib_uvs, img_sizes),
        total=n_cameras,
        ncols=72,
        disable=not verbose,
    ):
        intrinsics = get_intrinsics(
            uvs, calib_objpoints, img_size, n_samples=n_samples_for_intrinsics
        )
        all_intrinsics.append(intrinsics)

    all_calib_poses = []
    if verbose:
        print("Initializing calibration object poses")
    for uvs, intrinsics in tqdm.tqdm(
        zip(all_calib_uvs, all_intrinsics),
        total=n_cameras,
        disable=not verbose,
        ncols=72,
    ):
        calib_poses = estimate_pose(uvs, calib_objpoints, *intrinsics)
        all_calib_poses.append(calib_poses)
    all_calib_poses = np.array(all_calib_poses)

    if verbose:
        print("Estimating camera extrinsics")
    all_extrinsics = estimate_all_extrinsics(all_calib_poses, root=root)

    if verbose:
        print("Merging calibration object poses")
    calib_poses = consensus_calib_poses(all_calib_poses, all_extrinsics)

    return all_extrinsics, all_intrinsics, calib_poses
